
/******************************************************************************
* MODULE     : matrix_naive.hpp
* DESCRIPTION: Naive matrix operations
* COPYRIGHT  : (C) 2007  J. der Hoeven and G. Lecerf and O. Ruatta
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MATRIX_NAIVE__HPP
#define __MMX__MATRIX_NAIVE__HPP
#include <basix/vector.hpp>

namespace mmx {
#define TMPL template<typename C>

/******************************************************************************
* Matrix abstractions
******************************************************************************/

template<typename V>
struct matrix_naive_abstractions {
typedef V Vec;

template<typename Op, typename T, typename C> static inline void
mat_unary_stride (T* dest, nat dest_rs, nat dest_cs,
		  const C* src, nat src_rs, nat src_cs,
		  nat rows, nat cols) {
  for (; cols != 0; dest += dest_cs, src += src_cs, cols--)
    Vec::template vec_unary_stride<Op> (dest, dest_rs, src, src_rs, rows); }

template<typename Op, typename T, typename C1, typename C2> static inline void
mat_binary_stride (T* dest, nat dest_rs, nat dest_cs,
		   const C1* src1, nat src1_rs, nat src1_cs,
		   const C2* src2, nat src2_rs, nat src2_cs,
		   nat rows, nat cols) {
  for (; cols != 0; dest += dest_cs, src1 += src1_cs, src2 += src2_cs, cols--)
    Vec::template vec_binary_stride<Op>
      (dest, dest_rs, src1, src1_rs, src2, src2_rs, rows); }
};

/******************************************************************************
* Vectorial operations
******************************************************************************/

template<typename V>
struct matrix_vectorial: public V {
typedef typename V::Vec Vec;

TMPL static inline nat
vec_aligned_size (nat n) {
  return Vec::template vec_aligned_size<C> (n); }

static inline nat
index (nat row, nat col, nat rows, nat cols) {
  (void) cols;
  return col * rows + row; }

TMPL
static inline const C&
entry (const C* m, nat row, nat col, nat rows, nat cols) {
  return * (m + index (row, col, rows, cols)); }

/*
static inline nat
index (nat row, nat col, nat rows, nat cols) {
  (void) rows;
  return row * cols + col; }
*/

TMPL static void
set (C* dest, const C& c, nat rows, nat cols) {
  Vec::clear (dest, rows*cols);
  nat m= min (rows, cols), plus= index (1, 1, rows, cols);
  while (m != 0) {
    *dest= c;
    dest += plus; m--; } }

TMPL static void
extract (C* dest, const C* s,
	 nat rows, nat r_start, nat r_end,
	 nat cols, nat c_start, nat c_end) {
  nat rr= r_end - r_start, cc= c_end - c_start;
  V::template mat_unary_stride<id_op>
    (dest, index (1, 0, rr, cc), index (0, 1, rr, cc),
     s + index (r_start, c_start, rows, cols),
     index (1, 0, rows, cols), index (0, 1, rows, cols), rr, cc); }

TMPL static void
copy (C* dest, const C* s, nat r, nat c) {
  extract (dest, s, r, 0, r, c, 0, c); }

TMPL static void
transpose (C* dest, const C* src, nat rows, nat cols) {
  V::template mat_unary_stride<id_op>
    (dest, index (0, 1, cols, rows), index (1, 0, cols, rows),
     src,  index (1, 0, rows, cols), index (0, 1, rows, cols),
     rows, cols); }

}; // matrix_vectorial<V>

/******************************************************************************
* Operations on rows and columns
******************************************************************************/

template<typename V>
struct matrix_naive_row_column: public V {
typedef typename V::Vec Vec;

TMPL static inline void
vec_swap (C* s1, nat s1_stride,
	  C* s2, nat s2_stride,
	  nat n) {
  for (; n != 0; s1 += s1_stride, s2 += s2_stride, n--)
    swap (*s1, *s2); }

TMPL static void
row_swap (C* s, nat i, nat j, nat r, nat c) {
  ASSERT (i<r && j<r, "out of range");
  vec_swap (s + V::index (i, 0, r, c), V::index (0, 1, r, c),
	    s + V::index (j, 0, r, c), V::index (0, 1, r, c), c); }

TMPL static void
col_swap (C* s, nat i, nat j, nat r, nat c) {
  ASSERT (i<c && j<c, "out of range");
  vec_swap (s + V::index (0, i, r, c), V::index (1, 0, r, c),
	    s + V::index (0, j, r, c), V::index (1, 0, r, c), r); }

template<typename Op, typename C, typename X> static void
row_scalar (C* s, const X& x, nat i, nat r, nat c) {
  ASSERT (i<r, "out of range");
  Vec::template vec_unary_scalar_stride<Op>
    (s + V::index (i, 0, r, c), V::index (0, 1, r, c), x, c); }

template<typename Op, typename C, typename X> static void
col_scalar (C* s, const X& x, nat i, nat r, nat c) {
  ASSERT (i<c, "out of range");
  Vec::template vec_unary_scalar_stride<Op>
    (s + V::index (0, i, r, c), V::index (1, 0, r, c), x, r); }

TMPL static void row_mul (C* s, const C& sc, nat i, nat r, nat c) {
  row_scalar<rmul_op> (s, sc, i, r, c); }
TMPL static void row_div (C* s, const C& sc, nat i, nat r, nat c) {
  row_scalar<rdiv_op> (s, sc, i, r, c); }
TMPL static void col_mul (C* s, const C& sc, nat i, nat r, nat c) {
  col_scalar<rmul_op> (s, sc, i, r, c); }
TMPL static void col_div (C* s, const C& sc, nat i, nat r, nat c) {
  col_scalar<rdiv_op> (s, sc, i, r, c); }

TMPL static void
row_combine_sub (C* s, nat i, const C& si, nat j, const C& sj, nat r, nat c) {
  ASSERT (i<r && j<r, "out of range");
  C* iti= s + V::index (i, 0, r, c);
  C* itj= s + V::index (j, 0, r, c);
  nat plus= V::index (0, 1, r, c);
  while (c>0) {
    *iti= si * (*iti) - sj * (*itj);
    iti += plus; itj += plus; c--; } }

TMPL static void
col_combine_sub (C* s, nat i, const C& si, nat j, const C& sj, nat r, nat c) {
  ASSERT (i<c && j<c, "out of range");
  C* iti= s + V::index (0, i, r, c);
  C* itj= s + V::index (0, j, r, c);
  nat plus= V::index (1, 0, r, c);
  while (r>0) {
    *iti= si * (*iti) - sj * (*itj);
    iti += plus; itj += plus; r--; } }

TMPL static bool
row_is_zero (const C* s, nat i, nat r, nat c) {
  return Vec::template vec_binary_test_scalar_stride<equal_op>
    (s + V::index (i, 0, r, c), V::index (0, 1, r, c), C (0), c); }

TMPL static bool
col_is_zero (const C* s, nat j, nat r, nat c) {
  return Vec::template vec_binary_test_scalar_stride<equal_op>
    (s + V::index (0, j, r, c), V::index (1, 0, r, c), C (0), r); }

}; // matrix_naive_row_column<V>

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename V>
struct matrix_naive_multiplication: public V {
  typedef matrix_naive_multiplication<V> Naive;
  typedef matrix_naive_multiplication<V> Positive;
  typedef matrix_naive_multiplication<V> No_simd;
  typedef matrix_naive_multiplication<V> No_thread;
  typedef matrix_naive_multiplication<V> No_scaled;

template<typename Op, typename D, typename S1, typename S2>
static inline void
mul (D* d, const S1* s1, const S2* s2,
     nat r, nat rr, nat l, nat ll, nat c, nat cc) {
  typedef typename Op::acc_op Acc;
  if (l == 0) return;
  nat ri, ci, li;
  D *dr, *dc;
  const S1 *s1r, *s1c;
  const S2 *s2r, *s2c;
  nat drp = V::index (1, 0, rr, cc), dcp = V::index (0, 1, rr, cc);
  nat s1rp= V::index (1, 0, rr, ll), s1cp= V::index (0, 1, rr, ll);
  nat s2rp= V::index (1, 0, ll, cc), s2cp= V::index (0, 1, ll, cc);
  for (ri= r, dr= d, s1r= s1; ri!=0; ri--, dr += drp, s1r += s1rp)
    for (ci= c, dc= dr, s2c= s2; ci!=0; ci--, dc += dcp, s2c += s2cp) {
      D tmp= *dc;
      for (li= l, s1c= s1r, s2r= s2c; li==l; li--, s1c += s1cp, s2r += s2rp)
	Op ::in_place (tmp, *s1c, *s2r);
      for (                         ; li!=0; li--, s1c += s1cp, s2r += s2rp)
	Acc::in_place (tmp, *s1c, *s2r);
      *dc= tmp;
    } }
}; // matrix_naive_multiplication<V>

template<typename V>
struct matrix_enrich: public V {
  typedef matrix_enrich<typename V::Naive> Naive;
  typedef matrix_enrich<typename V::Positive> Positive;
  typedef matrix_enrich<typename V::No_simd> No_simd;
  typedef matrix_enrich<typename V::No_thread> No_thread;
  typedef matrix_enrich<typename V::No_scaled> No_scaled;

template<typename Op, typename D, typename S1, typename S2>
static inline void
mul (D* d, const S1* s1, const S2* s2,
     nat r, nat rr, nat l, nat ll, nat c, nat cc) {
  V::template mul<Op> (d, s1, s2, r, rr, l, ll, c, cc); }

template<typename D, typename S1, typename S2> static inline void
mul (D* dest, const S1* m1, const S2* m2, nat r, nat l, nat c) {
  V::template mul<mul_op> (dest, m1, m2, r, r, l, l, c, c); }

TMPL static inline void
print (ostream& out, const C* s, nat rows, nat cols) {
  out << "[";
  for (nat i=0; i < rows; i++) {
    for (nat j=0; j < cols; j++) {
      out << * (s + V::index (i, j, rows, cols));
      if (j + 1 != cols) out << ", ";
    }
    if (i + 1 != rows) out << ";\n ";
  }
  out << "]"; }
}; // matrix_enrich<V>

/******************************************************************************
* Matrix product iteration
******************************************************************************/

template<typename V>
struct matrix_naive_iterate: V {

TMPL static inline void
iterate_mul (C** d, const C* s, const C* x,
	     nat rs, nat cs, nat rx, nat cx, nat n) {
  // return s^0 x, s^1 x,..., s^{n-1} x in d
  // d must have allocated size n * rs * cx
  VERIFY (cs == rx && rs == cs, "sizes do not match");
  if (rs == 0) return;
  V::extract (d[0], x, rx, 0, rx, cx, 0, cx);
  for (nat i = 1; i < n; i++)
    V::mul (d[i], s, d[i-1], rs, cs, cx); }

TMPL static inline void
project_iterate_mul (C** d, const C* y, const C* s, const C* x,
		     nat ry, nat cy, nat rs, nat cs, nat rx, nat cx, nat n) {
  // return y s^0 x, y s^1 x,..., y s^{n-1} x in d
  VERIFY (cy == rs && rs == cs && cs == rx, "sizes do not match");
  if (n == 0) return;
  nat l= aligned_size<C,V> (rx * cx);
  C* t1= mmx_new<C> (l), * t2= mmx_new<C> (l);
  V::extract (t1, x, rx, 0, rx, cx, 0, cx);
  V::mul (d[0], y, t1, ry, cy, cx);
  for (nat i = 1; i < n; i++) {
    V::mul (t2, s, t1, rs, cs, cx);
    V::mul (d[i], y, t2, ry, cy, cx);
    swap (t1, t2);
  }
  mmx_delete<C> (t1, l); mmx_delete<C> (t2, l); }

}; // matrix_naive_iterate<V>

/******************************************************************************
* Determinant and characteristic polynomial with no division
******************************************************************************/
  
template<typename V>
struct matrix_naive_det_no_div: V {
private:

TMPL
static inline void
berkowitz (C* d, const C* s, nat rows, nat cols) {
  // x^rows + d[rows-1] x^[rows-1] + ... + d[0] is the characteristic
  // polynomial of s.
  // d must have allocated size at least rows + 1, and we set d[rows]=1
  // This is an implementation of Berkowitz' algorithm (1984).
  VERIFY (rows == cols, "square matrix expected");
  VERIFY (rows > 0, "non empty matrix expected");
  d[0]= - *(s + V::index (0, 0, rows, cols)); d[1]= 1;
  if (rows == 1) return;
  nat len_t= aligned_size<C,V> (rows + 1);
  C** t= mmx_new<C*> (len_t);
  for (nat j = 0; j < len_t; j++) t[j]= mmx_new<C> (1);
  C* t1= mmx_new<C> (len_t), * t2= mmx_new<C> (len_t);
  nat len_s= aligned_size<C,V> ((rows-1) * (cols-1));
  C* loc_s= mmx_new<C> (len_s);
  nat l= aligned_size<C,V> (rows-1);
  C* x= mmx_new<C> (l), * y= mmx_new<C> (l);
  for (nat i = 1; i < rows; i++) {
    // Apply Samuelson's formula to deduce charpoly (s[0..i, 0..i])
    // from charpoly (s[0..i-1, 0..i-1]).
    V::extract (loc_s, s, rows, 0, i, cols, 0, i);
    V::extract (y, s, rows, i, i+1, cols, 0, i  );
    V::extract (x, s, rows, 0, i  , cols, i, i+1);
    project_iterate_mul (t, y, loc_s, x, 1, i, i, i, i, 1, i);
    for (nat j = 0; j < i; j++) t1[j] = t[j][0];
    C a= * (s + V::index (i, i, rows, cols));
    V::mul (t2, t1, d + 1, 1, i, 1);  
    t2[0] += a * d[0];
    for (nat j = 1; j < i; j++) {
      V::mul (t2 + j, t1, d + j + 1, 1, i - j, 1);  
      t2[j] += a * d[j] - d[j-1];
    }
    t2[i]= a * d[i] - d[i-1];
    t2[i+1]= - d[i]; // namely -1
    V::Vec::neg (d, t2, i+2);
  }
  for (nat j = 0; j < len_t; j++) mmx_delete<C> (t[j], 1);
  mmx_delete<C*> (t, len_t);
  mmx_delete<C> (t1, len_t); mmx_delete<C> (t2, len_t);
  mmx_delete<C> (x, l); mmx_delete<C> (y, l);
  mmx_delete<C> (loc_s, len_s); }

public:

TMPL static inline C
det_no_div (const C* s, nat rows, nat cols) {
  VERIFY (rows == cols, "square matrix expected");
  C* d= mmx_new<C> (rows+1);
  berkowitz (d, s, rows, cols);
  C x= (rows & 1) ? - d[0] : d[0];
  mmx_delete<C> (d, rows+1);
  return x; }

TMPL static inline C
characteristic_polynomial_no_div (C* d, const C* s, nat rows, nat cols) {
  VERIFY (rows == cols, "square matrix expected");
  berkowitz (d, s, rows, cols); }

}; // matrix_naive_det_no_div

/******************************************************************************
* Echelon forms
******************************************************************************/

template<typename V>
struct matrix_naive_echelon: V {
private:

TMPL static void
row_echelon (C* m, C& num, C& den, C* k,
	     nat ri, nat ci, nat rf, nat cf,
	     nat rm, nat cm) {
  // Perform row pivoting from (ri, ci) to (rf, cf).
  // In return m is echelonized and k is the transformation matrix,
  // so that the computed m is the right product by k of the input m.
  // If k is NULL then it is not computed. Otherwise k must
  // have allocated size at least rm * rm.
  // In return num / den corresponds to the determinant
  // for a regular matrix.
  //
  // Note that the first entries of each echelonized row is not
  // ensured to be one.
  VERIFY ((ri < rm) && (ci < cm) && (rf <= rm) && (cf <= cm), "out of range");
  num= C(1); den= C(1);
  // search pivot
  nat i= ri, j= ci;
  for (j= ci; j<cf; j++) {
    for (i= ri; i<rf; i++)
      if (V::entry (m, i, j, rm, cm) != 0) break;
    if (i<rf) break;
  }
  if (i<rf && j<cf) {
    // put pivot on the top
    if (i != ri) {
      if (k != NULL) V::row_swap (k, ri, i, rm, rm);
      V::row_swap (m, ri, i, rm, cm);
      num= -num;
    }
    ci= j;
    // row reduction
    C p= V::entry (m, ri, ci, rm, cm);
    num *= p;
    for (nat index = 1; index < rf-ri; index++) {
      den *= p;
      C t= V::entry (m, ri + index, ci, rm, cm);
      if (k != NULL) V::row_combine_sub (k, ri + index, p, ri, t, rm, rm);
      V::row_combine_sub (m, ri + index, p, ri, t, rm, cm);
    }
    // call the pivoting on the non triangular sub-matrix
    if (ri+1 < rf && ci+1 < cf) {
      C x, y;
      row_echelon (m, x, y, k, ri+1, ci+1, rf, cf, rm, cm);
      num *= x; den *=y;
      return;
    } } }

TMPL static void
col_echelon (C* m, C& num, C& den, C* k,
	     nat ri, nat ci, nat rf, nat cf,
	     nat rm, nat cm) {
  // Perform column pivoting from (ri, ci) to (rf, cf).
  // Comments are the same as in row_echelon.
  VERIFY ((ri < rm) && (ci < cm) && (rf <= rm) && (cf <= cm), "out of range");
  num= C(1); den= C(1);
  // search pivot
  nat i= ri, j= ci;
  for (i= ri; i<rf; i++) {
    for (j= ci; j<cf; j++)
      if (V::entry (m, i, j, rm, cm) != 0) break;
    if (j<cf) break;
  }
  if (i<rf && j<cf) {
    // put pivot on the left
    if (j != ci) {
      if (k != NULL) V::col_swap (k, ci, j, cm, cm);
      V::col_swap (m, ci, j, rm, cm);
      num= -num;
    }
    ri= i;
    // column reduction
    C p= V::entry (m, ri, ci, rm, cm);
    num *= p;
    for (nat index = 1; index < cf-ci; index++) {
      den *= p;
      C t= V::entry (m, ri, ci + index, rm, cm);
      if (k != NULL) V::col_combine_sub (k, ci + index, p, ci, t, cm, cm);
      V::col_combine_sub (m, ci + index, p, ci, t, rm, cm);
    }
    // call the pivoting on the non triangular sub-matrix
    if (ri+1 < rf && ci+1 < cf) {
      C x, y;
      col_echelon (m, x, y, k, ri+1, ci+1, rf, cf, rm, cm);
      num *= x; den *=y;
    } } }

public:

TMPL static inline void
row_echelon (C* m, C& num, C& den, C* k, nat rm, nat cm) {
  if (k != NULL) V::set (k, C(1), rm, rm);
  row_echelon (m, num, den, k, 0, 0, rm, cm, rm, cm); }

TMPL static inline void
row_echelon (C* m, C* k, nat rm, nat cm) {
  C num, den;
  if (k != NULL) V::set (k, C(1), rm, rm);
  row_echelon (m, num, den, k, 0, 0, rm, cm, rm, cm); }

TMPL static inline void
row_echelon (C* m, nat rm, nat cm) {
  C num, den; C* k= NULL;
  row_echelon (m, num, den, k, 0, 0, rm, cm, rm, cm); }

TMPL static inline void
col_echelon (C* m, C& num, C& den, C* k, nat rm, nat cm) {
  if (k != NULL) V::set (k, C(1), cm, cm);
  col_echelon (m, num, den, k, 0, 0, rm, cm, rm, cm); }

TMPL static inline void
col_echelon (C* m, C* k, nat rm, nat cm) {
  if (k != NULL) V::set (k, C(1), cm, cm);
  C num, den;
  col_echelon (m, num, den, k, 0, 0, rm, cm, rm, cm); }

TMPL static inline void
col_echelon (C* m, nat rm, nat cm) {
  C num, den; C* k= NULL;
  col_echelon (m, num, den, k, 0, 0, rm, cm, rm, cm); }
}; // matrix_naive_echelon
  
/******************************************************************************
* Determinant and characteristic polynomial with exact division
******************************************************************************/

template<typename V>
struct matrix_naive_det: V {

TMPL static C
det (const C* m, nat rm, nat cm) {
  VERIFY (rm == cm, "square matrix expected");
  VERIFY (rm > 0 && cm > 0, "unexpected empty matrix");
  if (rm == 0) return 1;
  nat len_c= aligned_size<C,V> (rm * cm);
  C* c= mmx_new<C> (len_c), * k= NULL;
  C num, den;
  V::copy (c, m, rm, cm);
  V::col_echelon (c, num, den, k, rm, cm);
  C d= col_is_zero (c, cm - 1, rm, cm) ? C(0) : (num / den);
  mmx_delete<C> (c, len_c);
  return d; }
};

template<typename V>
struct matrix_naive_kernel: V {

TMPL static nat
kernel (C* k, const C* m, nat rm, nat cm) {
  // return the dimension of the kernel
  VERIFY (rm > 0 && cm > 0, "unexpected empty matrix");
  nat i, j, dim, len_c= aligned_size<C,V> (rm * cm);
  C* c= mmx_new<C> (len_c);
  V::copy (c, m, rm, cm);
  V::col_echelon (c, k, rm, cm);
  for (i=0; i<cm; i++)
    if (V::col_is_zero (c, i, rm, cm)) break;
  mmx_delete<C> (c, len_c);
  dim= cm - i;
  if (dim < cm)
    for (j=0; j < dim; j++)
      V::col_swap (k, j, i+j, cm, cm);
    // Note that the next columns are not filled with zeros
  return dim; }
};

template<typename V>
struct matrix_naive_image: V {

TMPL static nat
image (C* k, const C* m, nat rm, nat cm) {
  // return the dimension of the image
  VERIFY (rm > 0 && cm > 0, "unexpected empty matrix");
  nat i, len_c= aligned_size<C,V> (rm * cm);
  C* c= mmx_new<C> (len_c);
  V::copy (c, m, rm, cm);
  V::col_echelon (c, rm, cm);
  for (i=0; i<cm; i++)
    if (V::col_is_zero (c, i, rm, cm)) break;
  V::extract (k, c, rm, 0, rm, cm, 0, i);
  // Note that the next columns are not filled with zeros
  mmx_delete<C> (c, len_c);
  return i; }

TMPL static nat
rank (const C* m, nat rm, nat cm) {
  VERIFY (rm > 0 && cm > 0, "unexpected empty matrix");
  nat i, len_c= aligned_size<C,V> (rm * cm);
  C* c= mmx_new<C> (len_c);
  V::copy (c, m, rm, cm);
  V::col_echelon (c, rm, cm);
  for (i=0; i<cm; i++)
    if (V::col_is_zero (c, i, rm, cm)) break;
  mmx_delete<C> (c, len_c);
  return i; }
};

template<typename V>
struct matrix_naive_invert: V {

TMPL static void
invert (C* k, const C* m, nat rm, nat cm) {
  VERIFY (rm == cm, "square matrix expected");
  nat n= rm, l= aligned_size<C,V> (rm * cm);
  C* a= mmx_new<C> (l), * b= mmx_new<C> (l), * c= mmx_new<C> (l);
  V::copy (c, m, n, n);
  V::col_echelon (c, a, n, n); // m * a = c
  V::row_echelon (c, b, n, n); // b * c is diagonal
  for (nat i=0; i<n; i++) {
    C x= V::entry (c, i, i, n, n);
    ASSERT (x != 0, "non-invertible matrix");
    V::row_div (b, x, i, n, n);
  }
  V::mul (k, a, b, n, n, n);
  mmx_delete<C> (a, l); mmx_delete<C> (b, l); mmx_delete<C> (c, l); }
};

template<typename V>
struct matrix_naive_application: 
  public  matrix_naive_invert<
            matrix_naive_image<
	      matrix_naive_kernel<
		matrix_naive_det<
		  matrix_naive_echelon<
		    matrix_naive_det_no_div<
		      matrix_naive_iterate<V> > > > > > > {};

/******************************************************************************
* Naive matrices
******************************************************************************/

struct matrix_naive:
  public matrix_naive_application<
           matrix_enrich<
	     matrix_naive_multiplication<
	       matrix_naive_row_column<
		 matrix_vectorial<
		   matrix_naive_abstractions<
		     vector_naive> > > > > >
{
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

template<typename C>
struct matrix_variant_helper {
  typedef matrix_naive MV;
};

#undef TMPL
}//namespace mmx
#endif //__MMX__MATRIX_NAIVE__HPP
