
/******************************************************************************
* MODULE     : matrix_strassen.hpp
* DESCRIPTION: Matrix multiplication using Strassen's algorithm
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MATRIX_STRASSEN__HPP
#define __MMX__MATRIX_STRASSEN__HPP
#include <algebramix/matrix_simd.hpp>
#include <algebramix/matrix_threads.hpp>

namespace mmx {

/******************************************************************************
* Strassen multiplication (even-odd decomposition for numerical stability)
******************************************************************************/

template<nat thr, typename V>
struct matrix_strassen_multiplication: public V {
  typedef typename V::Naive Naive;
  typedef typename V::Positive Positive;
  typedef matrix_strassen_multiplication<thr,typename V::No_simd> No_simd;
  typedef matrix_strassen_multiplication<thr,typename V::No_thread> No_thread;
  typedef matrix_strassen_multiplication<thr,typename V::No_scaled> No_scaled;

  template<typename C> static inline void
  mat_load (C* d, nat r, nat c, const C* s, nat rr, nat cc) {
    nat j= c;
    for (; j!=0; j--, d += V::index(0,1,r,c), s += V::index(0,2,rr,cc)) {
      nat      i = r;
      C*       dd= d;
      const C* ss= s;
      for (; i!=0; i--, dd += V::index(1,0,r,c), ss += V::index(2,0,rr,cc))
	*dd= *ss;
    } 
  }

  template<typename Op, typename C> static inline void
  mat_save (C* d, nat rr, nat cc, const C* s, nat r, nat c) {
    typedef typename Op::set_op Set;
    nat j= c;
    for (; j!=0; j--, d += V::index(0,2,rr,cc), s += V::index(0,1,r,c)) {
      nat      i = r;
      C*       dd= d;
      const C* ss= s;
      for (; i!=0; i--, dd += V::index(2,0,rr,cc), ss += V::index(1,0,r,c))
	Set::in_place (*dd, *ss);
    } 
  }

  template<typename Op, typename D, typename S1, typename S2> static void
  mul (D* d, const S1* a, const S2* b,
       nat r, nat rr, nat l, nat ll, nat c, nat cc)
  {
    if (r < thr || l < thr || c < thr) {
      V::template mul<Op> (d, a, b, r, rr, l, ll, c, cc);
      return;
    }

    nat hr= r>>1, hl= l>>1, hc= c>>1, fr= hr<<1, fl= hl<<1, fc= hc<<1;
    
    S1* a11= mmx_new_safe<S1> (5 * hr * hl);
    S1* a12= a11 + hr * hl;
    S1* a21= a12 + hr * hl;
    S1* a22= a21 + hr * hl;
    S1* aaa= a22 + hr * hl;

    S2* b11= mmx_new_safe<S2> (5 * hl * hc);
    S2* b12= b11 + hl * hc;
    S2* b21= b12 + hl * hc;
    S2* b22= b21 + hl * hc;
    S2* bbb= b22 + hl * hc;

    D*  m1 = mmx_new_safe<D> (11 * hr * hc);
    D*  m2 = m1  + hr * hc;
    D*  m3 = m2  + hr * hc;
    D*  m4 = m3  + hr * hc;
    D*  m5 = m4  + hr * hc;
    D*  m6 = m5  + hr * hc;
    D*  m7 = m6  + hr * hc;
    D*  d11= m7  + hr * hc;
    D*  d12= d11 + hr * hc;
    D*  d21= d12 + hr * hc;
    D*  d22= d21 + hr * hc;

    mat_load (a11, hr, hl, a                          , rr, ll);
    mat_load (a12, hr, hl, a + V::index (0, 1, rr, ll), rr, ll);
    mat_load (a21, hr, hl, a + V::index (1, 0, rr, ll), rr, ll);
    mat_load (a22, hr, hl, a + V::index (1, 1, rr, ll), rr, ll);
    mat_load (b11, hr, hl, b                          , rr, ll);
    mat_load (b12, hr, hl, b + V::index (0, 1, rr, ll), rr, ll);
    mat_load (b21, hr, hl, b + V::index (1, 0, rr, ll), rr, ll);
    mat_load (b22, hr, hl, b + V::index (1, 1, rr, ll), rr, ll);

    V::Vec::add (aaa, a11, a22, hr * hl);
    V::Vec::add (bbb, b11, b22, hl * hc);
    mul<mul_op> (m1, aaa, bbb, hr, hr, hl, hl, hc, hc);
    V::Vec::add (aaa, a21, a22, hr * hl);
    mul<mul_op> (m2, aaa, b11, hr, hr, hl, hl, hc, hc);
    V::Vec::sub (bbb, b12, b22, hl * hc);
    mul<mul_op> (m3, a11, bbb, hr, hr, hl, hl, hc, hc);
    V::Vec::sub (bbb, b21, b11, hl * hc);
    mul<mul_op> (m4, a22, bbb, hr, hr, hl, hl, hc, hc);
    V::Vec::add (aaa, a11, a12, hr * hl);
    mul<mul_op> (m5, aaa, b22, hr, hr, hl, hl, hc, hc);
    V::Vec::sub (aaa, a21, a11, hr * hl);
    V::Vec::add (bbb, b11, b12, hl * hc);
    mul<mul_op> (m6, aaa, bbb, hr, hr, hl, hl, hc, hc);
    V::Vec::sub (aaa, a12, a22, hr * hl);
    V::Vec::add (bbb, b21, b22, hl * hc);
    mul<mul_op> (m7, aaa, bbb, hr, hr, hl, hl, hc, hc);

    V::Vec::add (d11, m1, m4, hr * hc);
    V::Vec::sub (d11, m5, hr * hc);
    V::Vec::add (d11, m7, hr * hc);
    V::Vec::add (d12, m3, m5, hr * hc);
    V::Vec::add (d21, m2, m4, hr * hc);
    V::Vec::sub (d22, m1, m2, hr * hc);
    V::Vec::add (d22, m3, hr * hc);
    V::Vec::add (d22, m6, hr * hc);

    mat_save<Op> (d                          , rr, cc, d11, hr, hc);
    mat_save<Op> (d + V::index (0, 1, rr, cc), rr, ll, d12, hr, hc);
    mat_save<Op> (d + V::index (1, 0, rr, cc), rr, ll, d21, hr, hc);
    mat_save<Op> (d + V::index (1, 1, rr, cc), rr, ll, d22, hr, hc);

    mmx_delete_safe<S1> (a11, 5  * hr * hl);
    mmx_delete_safe<S2> (b11, 5  * hl * hc);
    mmx_delete_safe<D > (m1 , 11 * hr * hc);

    mul_complete<Op,V> (d, a, b, r, rr, l, ll, c, cc, fr, fl, fc);
  }
};

/******************************************************************************
* Interface
******************************************************************************/

struct matrix_strassen:
  public matrix_naive_application<
           matrix_enrich<
             matrix_strassen_multiplication<512,
               matrix_simd_multiplication<
		 matrix_threads_multiplication<1024,4,
	           matrix_std_unroll_multiplication<
	             matrix_naive_multiplication<
		       matrix_naive_row_column<
			 matrix_vectorial<
			   matrix_naive_abstractions<
			     vector_naive> > > > > > > > > >
{
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

STMPL
struct matrix_variant_helper<double> {
  typedef matrix_strassen MV;
};

} // namespace mmx
#endif //__MMX__MATRIX_STRASSEN__HPP
