
/******************************************************************************
* MODULE     : polynomial_integer.hpp
* DESCRIPTION: Kronecker multiplication of integer polynomials
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_POLYNOMIAL_INTEGER_HPP
#define __MMX_POLYNOMIAL_INTEGER_HPP
#include <numerix/integer.hpp>
#include <algebramix/polynomial.hpp>
#include <algebramix/polynomial_dicho.hpp>
namespace mmx {

/******************************************************************************
* Low level Kronecker multiplication
******************************************************************************/

void mpz_kronecker_encode
       (mpz_t dest, mpz_t* src, unsigned long n, unsigned long bits);
void mpz_kronecker_decode
       (mpz_t* dest, unsigned long n, unsigned long bits, const mpz_t src_bis);

nat  max_bit_size (const integer* src, nat n);
void kronecker_encode (integer& dest, const integer* src, nat n, nat bits);
void kronecker_decode (integer* dest, nat n, nat bits, const integer& src);
void kronecker_mul (integer* dest,
		    const integer* src1, nat n1, const integer* src2, nat n2);
void kronecker_square (integer* dest, const integer* src1, nat n1);

/******************************************************************************
* Kronecker multiplication
******************************************************************************/

template<typename V, typename Th= dicho_threshold>
struct polynomial_kronecker_multiplication: public V {

template<typename C> static inline void
mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  V::mul (dest, s1, s2, n1, n2);
}

template<typename C> static inline void
square (C* dest, const C* s, nat n) {
  V::square (dest, s, n);
}

template<typename C> static inline void
tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  V::tmul (dest, s1, s2, n1, n2);
}

static inline void
mul (integer* dest, const integer* s1, const integer* s2, nat n1, nat n2) {
  if (min (n1, n2) < Threshold(integer,Th)) V::mul (dest, s1, s2, n1, n2);
  else kronecker_mul (dest, s1, n1, s2, n2);
}

static inline void
square (integer* dest, const integer* s, nat n) {
  if (n < Threshold(integer,Th)) V::square (dest, s, n);
  else kronecker_square (dest, s, n);
}

static inline void
tmul (integer* dest, const integer* s1, const integer* s2, nat n1, nat n2) {
  (void) dest; (void) s1; (void) s2; (void) n1; (void) n2;
  ERROR ("transposed Kronecker multiplication not implemented");
}

}; // polynomial_kronecker_multiplication<V>

template<typename V>
struct polynomial_multiply_integer:
  public polynomial_kronecker_multiplication<V>
{};

/******************************************************************************
* Use Kronecker multiplication by default for integer polynomials
******************************************************************************/

struct polynomial_integer_kronecker:
  public polynomial_algos_euclidean<
         polynomial_multiply_integer<
         polynomial_multiply_naive<
	   vector_naive> > >
{};

STMPL
struct polynomial_variant_helper<integer> {
  typedef Vector_variant(integer) VV;
  typedef polynomial_multiply_naive<VV> MN;
  typedef polynomial_multiply_integer<MN> MV;
  typedef polynomial_integer_kronecker PV;
  // typedef polynomial_algos_dicho<MV> PV;
  // Commented out so as to keep typenames simple when debugging
};

}
#endif // __MMX_POLYNOMIAL_INTEGER_HPP
