

/******************************************************************************
* MODULE     : subresultant_naive.hpp
* DESCRIPTION: Univariate subresultant of polynomials based on determinant
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_SUBRESULTANT_NAIVE_HPP
#define __MMX_SUBRESULTANT_NAIVE_HPP
#include <algebramix/matrix.hpp>
#include <algebramix/polynomial.hpp>

namespace mmx {

template<typename V>
struct subresultant_naive_algo: V {

#define TMPL       template<typename C>
#define Polynomial polynomial<C, typename V::PV>
#define Matrix     matrix    <C, typename V::MV>
#define PVector    vector    <Polynomial, typename V::VV>
private:
TMPL static Matrix
sylvester_matrix (const Polynomial& f, const Polynomial& g, int k, int l) {
  // If l = k then it returns the matrix of the following map:
  //   P_{dp - k} x P_{dq - k} --> P_{dp + dq - 2k}
  //                   (u, v) :--> (u * f + v * g) quo x^k,
  // in the canonical basis 1, x, x^2, etc.
  // In general, the determinant of the matrix returned is the coefficient
  // of degree l of the kth subresultant.
  int n = max (deg (f), 0), m = max (deg (g), 0), k2= 2*k;
  ASSERT (k >= 0 && k <= min (n, m), "out of range");
  ASSERT (l >= 0 && l <= k, "out of range");
  Matrix S (C (0), n + m - k2, n + m - k2);
  for (int j= 0; j < m - k; j++)
    S (0, j) = f[l - j]; 
  for (int j= 0; j < n - k; j++)
    S (0, j + m - k) = g[l - j]; 
  for (int j= 0; j < m - k; j++) 
    for (int i= 1; i < n + m - k2; i++)
      S (i, j) = f[k + i - j]; 
  for (int j= 0; j < n - k; j++)
    for (int i= 1; i < n + m - k2; i++)
      S (i, j + m - k) = g[k + i - j]; 
  return S; }

TMPL static inline Matrix
sylvester_matrix (const Polynomial& f, const Polynomial& g) {
  return sylvester_matrix (f, g, 0, 0); }

public:

TMPL static inline C
subresultant (const Polynomial& f, const Polynomial& g, int k, int l) {
  //! coefficient of degree l of the kth polynomial subresultant
  int m= min (deg (f), deg (g));
  if (m < 0 ||
      k < 0 || k >= m ||
      l < 0 || l > k) return C(0);
  return det (sylvester_matrix (f, g, k, l)); }

TMPL static Polynomial
subresultant (const Polynomial& f, const Polynomial& g, int k) {
  //! kth polynomial subresultant of f and g.
  // Only well-defined from degree 0 to m-1.
  Polynomial r(C(0));
  for (int l= 0; l <= k; l++)
    r += Polynomial (subresultant (f, g, k, l), l);
  return r; }

TMPL static PVector
subresultants (const Polynomial& f, const Polynomial& g) {
  //! Subresultants from degree 0 to min (deg (f), deg(g)) - 1.
  int m= min (deg (f), deg (g));
  if (m <= 0) return PVector (Polynomial (0), 0);
  PVector s (Polynomial (0), m);
  for (int k= 0; k < m; k++)
    s[k]= subresultant (f, g, k);
  return s; }

TMPL static inline C
resultant (const Polynomial& f, const Polynomial& g) {
  int m= min (deg (f), deg (g));
  if (m < 0) return C(0);
  return det (sylvester_matrix (f, g)); }
};

#undef TMPL
#undef Polynomial
#undef Matrix

/******************************************************************************
* Naive implementation of the subresultants
******************************************************************************/

struct subresultant_naive_base {
  typedef vector_naive     VV;
  typedef matrix_naive     MV;
  typedef polynomial_naive PV;
};

struct subresultant_naive :
    subresultant_naive_algo <subresultant_naive_base> {};

} // namespace mmx
#endif // __MMX_SUBRESULTANT_NAIVE_HPP
