
/******************************************************************************
* MODULE     : subresultant_prem.hpp
* DESCRIPTION: Univariate subresultant of polynomials based on pseudo division
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_SUBRESULTANT_PREM_HPP
#define __MMX_SUBRESULTANT_PREM_HPP
#include <algebramix/matrix.hpp>
#include <algebramix/polynomial.hpp>

namespace mmx {

#define TMPL       template<typename C>
#define Matrix     matrix    <C, typename V::MV>
#define Polynomial polynomial<C, typename V::PV>
#define PVector    vector    <Polynomial, typename V::VV>

template<typename V>
struct subresultant_prem_enrich: V {

TMPL static inline PVector
subresultants (const Polynomial& f, const Polynomial& g) {
  //! Subresultants from degree 0 to min (deg (f), deg(g)) - 1.
  int n= deg (f), m= deg (g);
  if (n < 0 || m < 0) return PVector (Polynomial (0), 0);
  if (n < m) {
    PVector s= V::subresultants (g, f);
    for (int k=0; k < n; k++) 
      s[k]= ((m+n-1) & (n-k) & 1) ? -s[k] : s[k];
    return s;
  }
  return V::subresultants (f, g); }

TMPL static Polynomial
subresultant (const Polynomial& f, const Polynomial& g, int k) {
  //! kth polynomial subresultant of f and g
  int m= min (deg (f), deg (g));
  if (m < 0 || k < 0 || k >= m) return Polynomial (C(0));
  PVector s= subresultants (f, g); return s[k]; }

TMPL static inline C
subresultant (const Polynomial& f, const Polynomial& g, int k, int l) {
  //! coefficient of degree l of the kth polynomial subresultant
  return subresultant (f, g, k)[l]; }

TMPL static inline C
resultant (const Polynomial& f, const Polynomial& g) {
  int n= deg (f), m= deg (g);
  if (n < 0 || m < 0) return C(0);
  if (m == 0) return binpow (g[0], n);
  if (n == 0) {
    C r= binpow (f[0], m); 
    return (n & 1) ? -r : r;
  }
  return subresultant (f, g, 0, 0); }
};

template<typename V>
struct subresultant_prem_algo: V {
  //! Lionel Ducos, "Optimizations of the subresultant algorithm,
  //! JPAA, 1998 (p. 150).

TMPL static PVector
subresultants (const Polynomial& f, const Polynomial& g) {
  //! Subresultants from degree 0 to min (deg (f), deg(g)) - 1.
  int n= deg (f), m= deg (g);
  VERIFY (n >= m, "incorrect degrees");
  PVector s (Polynomial (0), m);
  if (m == 0) return s;
  C d= binpow (g[m], n-m);
  Polynomial a= g, b= prem (f, g);
  s[m-1]= b;
  n= deg (a); m= deg (b);
  while (m >= 0) {
    s[m]= binpow (b[m], n-m-1) * b / binpow (d, n-m-1);
    if (m == 0) break;
    s[m-1]= prem (a, b) / (-a[n] * binpow (-d, n-m));
    a= s[m]; b= s[m-1]; d= a[m];
    n= deg (a); m= deg (b);
  }
  return s; }

};

template<typename V>
struct subresultant_optimized_prem_algo: V {
  //! Ibid, p. 154.
private:

TMPL static inline C
defected_pow (const C& x, const C& y, int n) {
  // return x^n / y^(n-1)
  VERIFY (n >= 1, "bug");
  if (n == 1) return x;
  C z= square (defected_pow (x, y, n >> 1)) / y;
  return ((n & 1) == 0) ? z : ((x * z) / y); }

TMPL static inline Polynomial
defected_scaling (const C& x, const C& y, const Polynomial& b, int n) {
  // return (x^n * b) / y^n
  VERIFY (n >= 0, "bug");
  if (n == 0) return b;
  if (n == 1) return (x * b) / y;
  return (defected_pow (x, y, n) * b) / y; }

TMPL static inline Polynomial
defected_prem (const Polynomial& s_d, const Polynomial& s_dm1,
	       const C& c_l, const C& x) {
  // return s_lm1 via c_l c_dm1 s_d = A s_dm1 + (-1)^(d-l+1) c_d^2 s_lm1
  // Ibid. p. 152.
  C c_dm1= s_dm1 [deg (s_dm1)];
  return rem (c_l * c_dm1 * s_d, s_dm1) / x; }

public:

TMPL static PVector
subresultants (const Polynomial& f, const Polynomial& g) {
  int n= deg (f), m= deg (g);
  VERIFY (n >= m, "incorrect degrees");
  PVector s (Polynomial (0), m);
  if (m == 0) return s;
  C d= binpow (g[m], n-m);
  Polynomial a= g, b= prem (f, g);
  s[m-1]= b;
  n= deg (a); m= deg (b);
  while (m >= 0) {
    s[m]= defected_scaling (b[m], d, b, n-m-1);
    if (m == 0) break;
    s[m-1]= defected_prem (a, b, s[m][deg (s[m])], d) / a[n];
    s[m-1]= ((n-m+1) & 1) ? -s[m-1] : s[m-1];
    a= s[m]; b= s[m-1]; d= a[m];
    n= deg (a); m= deg (b);
  }
  return s; }

};

#undef TMPL
#undef Matrix
#undef Polynomial
#undef PVector

/******************************************************************************
* Implementation of the subresultants via pseudo division
******************************************************************************/

struct subresultant_prem_base {
  typedef vector_naive     VV;
  typedef matrix_naive     MV;
  typedef polynomial_naive PV;
};

struct subresultant_prem :
    subresultant_prem_enrich <
      subresultant_prem_algo <subresultant_prem_base> > {};

struct subresultant_optimized_prem :
    subresultant_prem_enrich <
      subresultant_optimized_prem_algo <subresultant_prem_base> > {};

} // namespace mmx
#endif // __MMX_SUBRESULTANT_PREM_HPP
