
/******************************************************************************
* MODULE     : vector_unroll.hpp
* DESCRIPTION: loop unrolling within vectorial operations
* COPYRIGHT  : (C) 2007  Joris van der Hoeven and Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__VECTOR_UNROLL__HPP
#define __MMX__VECTOR_UNROLL__HPP
#include <algebramix/vector_fixed.hpp>

namespace mmx {
#define TMPL  template<typename C>

/******************************************************************************
* Abstract low level vector routines on vectors in memory
******************************************************************************/

template<nat len, typename V= vector_naive_abstractions>
struct vector_unroll_abstractions: public V {

typedef vector_unroll_abstractions<len, typename V::Aligned> Aligned;

template<typename Op, typename T> static inline void
vec_nullary (T* dest, nat n) {
loop:
  if (n<len) V::template vec_nullary<Op> (dest, n);
  else {
    vec_nullary_helper<Op, T, len>::op (dest);
    dest += len; n -= len;
    goto loop;
  }
}

template<typename Op, typename T, typename C> static inline void
vec_unary (T* dest, const C* s, nat n) {
loop:
  if (n<len) V::template vec_unary<Op> (dest, s, n);
  else {
    vec_unary_helper<Op, T, C, len>::op (dest, s);
    dest += len; s += len; n -= len;
    goto loop;
  }
}

template<typename Op, typename T, typename C1, typename C2> static inline void
vec_binary (T* dest, const C1* s1, const C2* s2, nat n) {
loop:
  if (n<len) V::template vec_binary<Op, T, C1, C2> (dest, s1, s2, n);
  else {
    vec_binary_helper<Op, T, C1, C2, len>::op (dest, s1, s2);
    dest += len; s1 += len; s2 += len; n -= len;
    goto loop;
  }
}

template<typename Op, typename T, typename X> static inline void
vec_unary_scalar (T* dest, const X& x, nat n) {
loop:
  if (n<len) V::template vec_unary_scalar<Op, T, X> (dest, x, n);
  else {
    vec_unary_scalar_helper<Op, T, X, len>::op (dest, x);
    dest += len; n -= len;
    goto loop;
  }
}

template<typename Op, typename T, typename C, typename X> static inline void
vec_binary_scalar (T* dest, const C* s, const X& x, nat n) {
loop:
  if (n<len) V::template vec_binary_scalar<Op, T, C, X> (dest, s, x, n);
  else {
    vec_binary_scalar_helper<Op, T, C, X, len>::op (dest, s, x);
    dest += len; s += len; n -= len;
    goto loop;
  }
}

template<typename Op, typename C> static inline bool
vec_unary_test (const C* s, nat n) {
loop:
  if (n<len) 
    return V::template vec_unary_test<Op, C> (s, n);
  else {
    if (! vec_unary_test_helper<Op, C, len>::op (s))
      return false;
    s += len; n -= len;
    goto loop;
  }
  return true;
}

template<typename Op, typename C1, typename C2> static inline bool
vec_binary_test (const C1* s1, const C2* s2, nat n) {
loop:
  if (n<len) 
    return V::template vec_binary_test<Op, C1, C2> (s1, s2, n);
  else {
    if (! vec_binary_test_helper<Op, C1, C2, len>::op (s1, s2))
      return false;
    s1 += len; s2 += len; n -= len;
    goto loop;
  }
  return true;
}

template<typename Op, typename C, typename X> static inline bool
vec_binary_test_scalar (const C* s, const X& x, nat n) {
loop:
  if (n<len)
    return V::template vec_binary_test_scalar<Op, C, X> (s, x, n);
  else {
    if (! vec_binary_test_scalar_helper<Op, C, X, len>::op (s, x))
      return false;
    s += len; n -= len;
    goto loop;
  }
  return true;
}

template<typename Op, typename C> static inline C
vec_unary_big (const C* s, nat n) {
  if (n == 0) return Op::template neutral<C> ();
  C r= *s;
  s++; n--;
loop:
  if (n < len)
    Op::in_place (r, V::template vec_unary_big<Op, C> (s, n));
  else {
    Op::in_place (r, vec_unary_big_helper<Op, C, len>::op (s));
    s += len; n -= len;
    goto loop;
  }
  return r;
}

template<typename Op, typename C> static inline C
vec_unary_big_dicho (const C* s, nat n) {
  return vec_unary_big<Op, C> (s, n);
}

template<typename Op, typename C1, typename C2> static inline C1
vec_binary_big (const C1* s1, const C2* s2, nat n) {
  C1 r= Op::template neutral<C1> ();
loop:
  if (n<len)
    for (; n != 0; s1++, s2++, n--)
      Op::in_place (r, *s1, *s2);
  else {
    vec_binary_big_helper<Op, C1, C2, len>::_op (r, s1, s2);
    s1 += len; s2 += len; n -= len;
    goto loop;
  }
  return r;
}

}; // vector_unroll_abstractions

template<nat len1, nat len2 = len1>
struct vector_unroll:
    public vector_enrich<
  vector_unroll_abstractions<len2, vector_unroll_abstractions<len1> > >
{
  static const nat def_len = 0;
  static const nat init_len= 1;
};

/******************************************************************************
* Use unrolled vector operations by default
******************************************************************************/

STMPL
struct vector_variant_helper<double> {
  typedef vector_unroll<8> VV;
};

STMPL
struct vector_variant_helper<char> {
  typedef vector_unroll<32> VV;
};

STMPL
struct vector_variant_helper<signed char> {
  typedef vector_unroll<32> VV;
};

STMPL
struct vector_variant_helper<unsigned char> {
  typedef vector_unroll<32> VV;
};

STMPL
struct vector_variant_helper<short int> {
  typedef vector_unroll<32> VV;
};

STMPL
struct vector_variant_helper<short unsigned int> {
  typedef vector_unroll<32> VV;
};

STMPL
struct vector_variant_helper<int> {
  typedef vector_unroll<16> VV;
};

STMPL
struct vector_variant_helper<unsigned int> {
  typedef vector_unroll<16> VV;
};

STMPL
struct vector_variant_helper<long int> {
  typedef vector_unroll<16> VV;
};

STMPL
struct vector_variant_helper<long unsigned int> {
  typedef vector_unroll<16> VV;
};

STMPL
struct vector_variant_helper<long long int> {
  typedef vector_unroll<8> VV;
};

STMPL
struct vector_variant_helper<long long unsigned int> {
  typedef vector_unroll<8> VV;
};

#undef TMPL
} // namespace mmx
#endif //__MMX__VECTOR_UNROLL__HPP
