
/******************************************************************************
* MODULE     : fast_new.hpp
* DESCRIPTION: Fast memory allocation and new operation
* COPYRIGHT  : (C) 1998-2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_FAST_NEW_HPP
#define __MMX_FAST_NEW_HPP
#include <basix/basix-config.hpp>
#include <iostream>
#include <cassert>

#include <stdlib.h>
#ifdef BASIX_OLD_GNU_COMPILER
inline void* operator new (register size_t s, void* loc) { return loc; }
inline void* operator new[] (register size_t s, void* loc) { return loc; }
#else
#include <new>
#endif

using std::istream;
using std::ostream;
using std::cin;
using std::cout;
using std::cerr;

#define ENABLE_FAST_ALLOCATORS

//! \file

/*
#ifdef USE_STANDARD_NEW_DELETE
// Optional check that standard new and delete are never used
inline void* operator new (register size_t s) { assert (false); }
inline void  operator delete (register void* ptr) { assert (false); }
inline void* operator new[] (register size_t s) { assert (false); }
inline void  operator delete[] (register void* ptr) { assert (false); }
#endif
*/

namespace mmx {

typedef unsigned int nat;
typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned long ulong;

/******************************************************************************
* Allocators and Deallocators
******************************************************************************/

void*  mmx_malloc (size_t new_size);
void*  mmx_realloc (void* ptr, size_t old_size, size_t new_size);
void   mmx_free (void* ptr, size_t old_size);
size_t mmx_used_bytes ();

#define MMX_ALLOCATORS						\
public:								\
  inline void* operator new (size_t sz) {			\
    return mmx_malloc (sz); }					\
  inline void* operator new (size_t sz, void* where) {		\
    (void) sz; return where; }					\
  inline void* operator new[] (size_t sz) {			\
    return mmx_malloc (sz); }					\
  inline void* operator new[] (size_t sz, void* where) {	\
    (void) sz; return where; }					\
  inline void operator delete (void* ptr, size_t sz) {		\
    mmx_free (ptr, sz); }					\
  inline void operator delete[] (void* ptr, size_t sz) {	\
    mmx_free (ptr, sz); }

/******************************************************************************
* Fast array allocators
******************************************************************************/

#ifdef ENABLE_FAST_ALLOCATORS

template<typename C> inline C*
mmx_new (nat n) {
#ifdef BASIX_ENABLE_VERIFY
  void* ptr= mmx_malloc (n * sizeof (C) + 16);
  *((nat*) ptr)= n;
  //cout << "{+" << ptr << "}"; cout.flush ();
  ptr= (void*) (((char*) ptr) + 16);
  //cout << "{++" << ptr << "}"; cout.flush ();
#else
  void* ptr= mmx_malloc (n * sizeof (C));
#endif
  C* ctr= (C*) ptr;
  for (nat i=0; i<n; i++, ctr++)
    (void) new ((void*) ctr) C ();
  return (C*) ptr;
}

template<typename C, typename T1> inline C*
mmx_new (nat n, const T1& a1) {
#ifdef BASIX_ENABLE_VERIFY
  void* ptr= mmx_malloc (n * sizeof (C) + 16);
  *((nat*) ptr)= n;
  //cout << "{.+" << ptr << "}"; cout.flush ();
  ptr= (void*) (((char*) ptr) + 16);
  //cout << "{.++" << ptr << "}"; cout.flush ();
#else
  void* ptr= mmx_malloc (n * sizeof (C));
#endif
  C* ctr= (C*) ptr;
  for (nat i=0; i<n; i++, ctr++)
    (void) new ((void*) ctr) C (a1);
  return (C*) ptr;
}

template<typename C, typename T1, typename T2> inline C*
mmx_new (nat n, const T1& a1, const T2& a2) {
#ifdef BASIX_ENABLE_VERIFY
  void* ptr= mmx_malloc (n * sizeof (C) + 16);
  *((nat*) ptr)= n;
  //cout << "{..+" << ptr << "}"; cout.flush ();
  ptr= (void*) (((char*) ptr) + 16);
  //cout << "{..++" << ptr << "}"; cout.flush ();
#else
  void* ptr= mmx_malloc (n * sizeof (C));
#endif
  C* ctr= (C*) ptr;
  for (nat i=0; i<n; i++, ctr++)
    (void) new ((void*) ctr) C (a1, a2);
  return (C*) ptr;
}

template<typename C> inline void
mmx_delete (C* Ptr, nat n) {
#ifdef BASIX_ENABLE_VERIFY
  void* ptr= (void*) Ptr;
  //cout << "{--" << ptr << "}"; cout.flush ();
  ptr= (void*) (((char*) ptr) - 16);
  //cout << "{-" << ptr << "}"; cout.flush ();
  if (*((nat*) ptr) != n)
    cout << "[" << n << " should be " << *((nat*) ptr) << "]\n";
  assert (*((nat*) ptr) == n);
  C* ctr= Ptr+n-1;
  for (nat i=0; i<n; i++, ctr--) ctr -> ~C();
  mmx_free (ptr, n * sizeof (C) + 16);
#else
  C* ctr= Ptr+n-1;
  for (nat i=0; i<n; i++, ctr--) ctr -> ~C();
  mmx_free ((void*) Ptr, n * sizeof (C));
#endif
}

template<typename C> inline C*
mmx_new_one () {
  return mmx_new<C> (1);
}

template<typename C, typename T1> inline C*
mmx_new_one (const T1& a1) {
  return mmx_new<C> (1, a1);
}

template<typename C> inline void
mmx_delete_one (C* ptr) {
  mmx_delete<C> (ptr, 1);
}

template<typename C> inline C*
mmx_classical_new (nat n) {
  void* ptr= mmx_malloc (n * sizeof (C) + sizeof (nat));
  *((nat*) ptr)= n;
  ptr= (void*) (((nat*) ptr) + 1);
  C* ctr= (C*) ptr;
  for (nat i=0; i<n; i++, ctr++)
    (void) new ((void*) ctr) C ();
  return (C*) ptr;
}

template<typename C> inline void
mmx_classical_delete (C* Ptr) {
  void* ptr= (void*) (((nat*) ((void*) Ptr)) - 1);
  nat n= *((nat*) ptr);
  C* ctr= Ptr+n-1;
  for (nat i=0; i<n; i++, ctr--) ctr -> ~C();
  mmx_free (ptr, n * sizeof (C) + sizeof (nat));
}

/******************************************************************************
* Wrapper for slower default allocators
******************************************************************************/

#else // not ENABLE_FAST_ALLOCATORS

template<typename C> inline C*
mmx_new (nat n) {
  return new C[n];
}

template<typename C, typename T1> inline C*
mmx_new (nat n, const T1& a1) {
  C* ptr= new C[n];
  for (nat i=0; i<n; i++) ptr[i]= C(a1);
  return ptr;
}

template<typename C> inline void
mmx_delete (C* ptr, nat n) {
  (void) n;
  delete[] ptr;
}

template<typename C> inline C*
mmx_new_one () {
  return new C ();
}

template<typename C, typename T1> inline C*
mmx_new_one (const T1& a1) {
  return new C (a1);}

template<typename C> inline void
mmx_delete_one (C* ptr) {
  delete ptr;
}

template<typename C> inline C*
mmx_classical_new (nat n) {
  return new C[n];
}

template<typename C> inline void
mmx_classical_delete (C* ptr) {
  delete[] ptr;
}

#endif // ENABLE_FAST_ALLOCATORS

} // namespace mmx
#endif // __MMX_FAST_NEW_HPP
