
/******************************************************************************
* MODULE     : threads.hpp
* DESCRIPTION: Multi-threading
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_THREADS_HPP
#define __MMX_THREADS_HPP
#include <basix/exception.hpp>

#ifdef BASIX_ENABLE_THREADS
#include <pthread.h>
#endif

//! \file

namespace mmx {

/******************************************************************************
* Parallel execution of tasks
******************************************************************************/

class task_rep REP_STRUCT {
public:
  inline task_rep () {}
  virtual ~task_rep () {}
  virtual void execute () = 0;
};

class task {
INDIRECT_PROTO (task, task_rep)
public:
  inline task (): rep (NULL) {}
};
INDIRECT_NULL_IMPL (task, task_rep)

extern bool threads_active;
extern nat  threads_number;
void threads_initialize (nat n);
void threads_terminate ();
void threads_execute (task* ts, nat n);
void threads_simple_loop (nat nr_instances, nat nr_iterations);

/******************************************************************************
* Mutexes
******************************************************************************/

#ifdef BASIX_ENABLE_THREADS

class mutex {
  pthread_mutex_t rep;
public:
  inline mutex () { pthread_mutex_init (&rep, NULL); }
  inline ~mutex () { pthread_mutex_destroy (&rep); }
  friend class mutex_lock;
};

class mutex_lock {
  pthread_mutex_t* rep;
public:
  inline mutex_lock (mutex& m): rep (&(m.rep)) {
    pthread_mutex_lock (rep); }
  inline ~mutex_lock () {
    pthread_mutex_unlock (rep); }
};

#else

class mutex {};
class mutex_lock { inline mutex_lock (mutex& m) { (void) m; } };

#endif

/******************************************************************************
* Thread-safe memory management
******************************************************************************/

//extern mutex memory_lock;

void* mmx_malloc_safe (size_t new_size);
void* mmx_realloc_safe (void* ptr, size_t old_size, size_t new_size);
void  mmx_free_safe (void* ptr, size_t old_size);

template<typename C> inline C*
mmx_new_safe (nat n) {
  //mutex_lock lock (memory_lock);
#ifdef BASIX_ENABLE_VERIFY
  void* ptr= mmx_malloc_safe (n * sizeof (C) + 16);
  *((nat*) ptr)= n;
  //cout << "{+" << ptr << "}"; cout.flush ();
  ptr= (void*) (((char*) ptr) + 16);
  //cout << "{++" << ptr << "}"; cout.flush ();
#else
  void* ptr= mmx_malloc_safe (n * sizeof (C));
#endif
  C* ctr= (C*) ptr;
  for (nat i=0; i<n; i++, ctr++)
    (void) new ((void*) ctr) C ();
  return (C*) ptr;
}

template<typename C> inline void
mmx_delete_safe (C* Ptr, nat n) {
  //mutex_lock lock (memory_lock);
#ifdef BASIX_ENABLE_VERIFY
  void* ptr= (void*) Ptr;
  //cout << "{--" << ptr << "}"; cout.flush ();
  ptr= (void*) (((char*) ptr) - 16);
  //cout << "{-" << ptr << "}"; cout.flush ();
  if (*((nat*) ptr) != n)
    cout << "[" << n << " should be " << *((nat*) ptr) << "]\n";
  assert (*((nat*) ptr) == n);
  C* ctr= Ptr+n-1;
  for (nat i=0; i<n; i++, ctr--) ctr -> ~C();
  mmx_free_safe (ptr, n * sizeof (C) + 16);
#else
  C* ctr= Ptr+n-1;
  for (nat i=0; i<n; i++, ctr--) ctr -> ~C();
  mmx_free_safe ((void*) Ptr, n * sizeof (C));
#endif
}

} // namespace mmx

#endif // __MMX_THREADS_HPP
