
/******************************************************************************
* MODULE     : vector.hpp
* DESCRIPTION: Dense univariate vectors
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_VECTOR_HPP
#define __MMX_VECTOR_HPP
#include <basix/iterator.hpp>
#include <basix/vector_naive.hpp>

//! \file

namespace mmx {
#define Vector_variant(C) vector_variant_helper<C>::VV
#define TMPL_DEF template<typename C, typename V>
// NOTE: default argument vector_variant_helper<C>::VV specified in basix.hpp
#define TMPL template<typename C, typename V>
#define Vector vector<C,V>
#define Vector_int vector<int,V>
#define Vector_rep vector_rep<C,V>
#define Abs_vector vector<Abs_type(C),V>
TMPL class vector;
TMPL inline nat N (const Vector& v);
TMPL inline C* seg (Vector& v);
TMPL inline const C* seg (const Vector& v);
TMPL inline bool is_a_scalar   (const Vector& v);
TMPL inline bool is_non_scalar (const Vector& v);

/******************************************************************************
* Memory alignment
******************************************************************************/

template<typename C, typename V> inline nat
aligned_size (nat n) {
  return V::template vec_aligned_size<C> (n); }

template<typename C> inline nat
default_aligned_size (nat n) {
  typedef typename Vector_variant(C) V;
  return aligned_size<C,V> (n); }

/******************************************************************************
* Dense vectors of arbitrary sizes
******************************************************************************/

TMPL_DEF
class vector_rep REP_STRUCT {
  C*   a;           // entries of vector
  nat  n;           // dimension of vector
  nat  l;           // allocated number of entries
  bool scalar_flag; // is the vector a scalar value?
public:
  inline vector_rep (C* a2, nat n2, bool flag):
    a (a2), n (n2), l (n2), scalar_flag (flag) {}
  inline vector_rep (C* a2, nat n2, nat l2, bool flag):
    a (a2), n (n2), l (l2), scalar_flag (flag) {}
  inline ~vector_rep () { mmx_delete<C> (a, l); }
  void resize (nat n2) {
    nat l2;
    if (n2 > l) l2= aligned_size<C,V> (max (n2, l<<1));
    else if (n2 < (l >> 1)) l2= aligned_size<C,V> (n2);
    else { n= n2; return; }
    C* b= mmx_new<C> (l2);
    V::copy (b, a, min (n, n2));
    mmx_delete<C> (a, l);
    a= b;
    n= n2;
    l= l2;
  }
  friend class Vector;
  friend nat N LESSGTR (const Vector& v);
  friend C* seg LESSGTR (Vector& v);
  friend const C* seg LESSGTR (const Vector& v);
  friend bool is_a_scalar   LESSGTR (const Vector& v);
  friend bool is_non_scalar LESSGTR (const Vector& v);
};

TMPL_DEF
class vector {
INDIRECT_PROTO_2 (vector, vector_rep, C, V)
public:
  vector () {
    nat n= V::def_len;
    nat l= aligned_size<C,V> (n);
    rep= new Vector_rep (mmx_new<C> (l), n, l, false);
  }
  template<typename T> vector (const T& c) {
    nat n= V::init_len;
    nat l= aligned_size<C,V> (n);
    rep= new Vector_rep (mmx_new<C> (l), n, l, true);
    V::set (rep->a, C(c), n);
  }
  template<typename T, typename W> vector (const vector<T,W>& v) {
    nat l= aligned_size<C,V> (N(v));
    rep= new Vector_rep (mmx_new<C> (l), N(v), l, is_a_scalar (v));
    V::cast (rep->a, seg (v), N(v));
  }
  vector (const C& x) {
    nat n= V::init_len;
    nat l= aligned_size<C,V> (n);
    rep= new Vector_rep ( mmx_new<C> (l), n, l, true);
    V::set (rep->a, x, n);
  }
  vector (const C& x, nat n) {
    nat l= aligned_size<C,V> (n);
    rep= new Vector_rep (mmx_new<C> (l), n, l, false);
    V::set (rep->a, x, n);
  }
  inline vector (C* a, nat n=V::def_len) {
    rep= new Vector_rep (a, n, false);
  }
  inline vector (C* a, nat n, nat l) {
    rep= new Vector_rep (a, n, l, false);
  }
  inline vector (C* a, nat n, nat l, bool flag) {
    rep= new Vector_rep (a, n, l, flag);
  }
  vector (const iterator<C>& it) {
    nat i, l= V::def_len;
    if (l==0) l= aligned_size<C,V> ((nat) 1);
    C* a= mmx_new<C> (l);
    for (i=0; busy (it); i++, ++it) {
      if (i >= l) {
	C* b= mmx_new<C> (aligned_size<C,V> (l<<1));
	V::copy (b, a, l);
	mmx_delete (a, l);
	a= b;
	l= l<<1;
      }
      a[i]= *it;
    }
    rep= new Vector_rep (a, i, l, false);
  }

  inline const C& operator [] (nat i) const {
    VERIFY (is_non_scalar (*this), "non-scalar vector expected");
    VERIFY (i<N(*this), "index out of range");
    return rep->a[i]; }
  inline C& operator [] (nat i) {
    VERIFY (is_non_scalar (*this), "non-scalar vector expected");
    VERIFY (i<N(*this), "index out of range");
    secure(); return rep->a[i]; }
  inline C scalar () const {
    VERIFY (is_a_scalar (*this), "scalar vector expected");
    return rep->a[0]; }
  inline C& scalar () {
    VERIFY (is_a_scalar (*this), "scalar vector expected");
    return rep->a[0]; }

  Vector& operator << (const C& append);
  Vector& operator << (const Vector& w);
};
INDIRECT_IMPL_2 (vector, vector_rep, typename C, C, typename V, V)

TMPL inline nat N (const Vector& v) { return v->n; }
TMPL inline bool is_a_scalar (const Vector& v) { return v->scalar_flag; }
TMPL inline bool is_non_scalar (const Vector& v) { return !v->scalar_flag; }
TMPL inline const C& read (const Vector& v, nat i) { return v[i]; }
TMPL inline const C* seg (const Vector& v) { return v->a; }
TMPL inline C* seg (Vector& v) {
  VERIFY (is_non_scalar (v), "non-scalar vector expected");
  v.secure(); return v->a; }
TMPL inline Vector extend (const Vector& v, const Vector& w) {
  VERIFY (is_a_scalar (v), "scalar vector expected");
  VERIFY (is_non_scalar (w), "non-scalar vector expected");
  return Vector (v.scalar(), N(w)); }
TMPL inline bool is_nil (const Vector& v) {
  return N(v) == 0; }
TMPL inline bool is_atom (const Vector& v) {
  return is_non_scalar (v) && N(v) == 1; }

STYPE_TO_TYPE(TMPL,scalar_type,Vector,C);
STYPE_TO_TYPE(TMPL,abs_type,Vector,Abs_vector);
TMPL struct rounding_helper<Vector >: public rounding_helper<C> {};
TMPL struct species_type_information<Vector > {
  static const nat id= SPECIES_VECTOR; };

TMPL
struct fast_helper<Vector > {
  typedef Fast_type(C) FC;
  typedef vector<FC> fast_type;
  typedef typename Vector_variant(FC) FV;
  static inline fast_type dd (const Vector& v) {
    nat l= aligned_size<FC, FV> (N(v));
    FC* r= mmx_new<FC> (l);
    for (nat i=0; i<N(v); i++) r[i]= fast (v[i]);
    return fast_type (r, N(v), l, is_a_scalar (v)); }
   static inline Vector uu (const fast_type& v) {
     nat l= aligned_size<C,V> ((N(v)));
     C* r= mmx_new<C> (l);
     for (nat i=0; i<N(v); i++) r[i]= slow<C> (v[i]);
     return Vector (r, N(v), l, is_a_scalar (v)); }
 };

template<typename T, typename F, typename TV, typename FV>
struct as_helper<vector<T,TV>, vector<F,FV> > {
  static inline vector<T,TV>
  cv (const vector<F,FV>& v) {
    nat l= aligned_size<T,TV> (N(v));
    T* c= mmx_new<T> (l);
    for (nat i=0; i<N(v); i++) c[i]= as<T> (v[i]);
    return vector<T,TV> (c, N(v), l, is_a_scalar (v)); }
};

/******************************************************************************
* Fixed size dense vectors
******************************************************************************/

template<typename V, typename S>
struct vector_fixed: public V {
  static const nat def_len = S::val;
  static const nat init_len= S::val;
};

#define FVector     vector    <C,vector_fixed<V,S> >
#define FVector_rep vector_rep<C,vector_fixed<V,S> >

template<typename C, typename V, typename S>
class FVector_rep REP_STRUCT {
  C* a;
public:
  inline vector_rep (C* a2, nat n, bool dummy): a (a2) {
    VERIFY (n == S::val, "wrong size");
    (void) n; (void) dummy; }
  inline vector_rep (C* a2, nat n, nat l, bool dummy): a (a2) {
    VERIFY (n == S::val, "wrong size");
    VERIFY ((l == aligned_size<C,V> (S::val)), "wrong allocated size");
    (void) n; (void) l; (void) dummy; }
  inline virtual ~vector_rep () {
    mmx_delete<C> (a, aligned_size<C,V> (S::val)); }
  friend class FVector;
  friend nat N LESSGTR (const FVector& v);
  friend C*  seg LESSGTR (FVector& v);
  friend const C* seg LESSGTR (const FVector& v);
  friend bool is_a_scalar   LESSGTR (const Vector& v);
  friend bool is_non_scalar LESSGTR (const Vector& v);
};

template<typename C, typename V, typename S> inline nat
N (const FVector& v) { return S::val; }
template<typename C, typename V, typename S> inline bool
is_a_scalar (const FVector& v) { (void) v; return false; }
template<typename C, typename V, typename S> inline bool
is_non_scalar (const FVector& v) { (void) v; return true; }

#undef FVector
#undef FVector_rep

/******************************************************************************
* Explicit dense vectors
******************************************************************************/

template<typename C> vector<C>
vec () {
  nat l= default_aligned_size<C> ((nat) 0);
  C* a= mmx_new<C> (l);
  return vector<C> (a, 0, l);
}

template<typename C> vector<C>
vec (const C& c1) {
  nat l= default_aligned_size<C> ((nat) 1);
  C* a= mmx_new<C> (l);
  a[0]= c1;
  return vector<C> (a, 1, l);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2) {
  nat l= default_aligned_size<C> ((nat) 2);
  C* a= mmx_new<C> (l);
  a[0]= c1; a[1]= c2;
  return vector<C> (a, 2, l);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3) {
  nat l= default_aligned_size<C> ((nat) 3);
  C* a= mmx_new<C> (l);
  a[0]= c1; a[1]= c2; a[2]= c3;
  return vector<C> (a, 3, l);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3, const C& c4) {
  nat l= default_aligned_size<C> ((nat) 4);
  C* a= mmx_new<C> (l);
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4;
  return vector<C> (a, 4, l);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3, const C& c4, const C& c5) {
  nat l= default_aligned_size<C> ((nat) 5);
  C* a= mmx_new<C> (l);
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4; a[4]= c5;
  return vector<C> (a, 5, l);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3,
     const C& c4, const C& c5, const C& c6) {
  nat l= default_aligned_size<C> ((nat) 6);
  C* a= mmx_new<C> (l);
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4; a[4]= c5; a[5]= c6;
  return vector<C> (a, 6, l);
}

template<typename C> inline vector<C>
fill (nat n) {
  nat l= default_aligned_size<C> (n);
  return vector<C> (mmx_new<C> (l), n, l);
}

template<typename C> inline vector<C>
fill (const C& c, nat n) {
  nat l= default_aligned_size<C> (n);
  C* a= mmx_new<C> (l);
  for (nat i=0; i<n; i++) a[i]= c;
  return vector<C> (a, n, l);
}

/******************************************************************************
* Vector iteration
******************************************************************************/

TMPL_DEF
class vector_iterator_rep: public iterator_rep<C> {
  const Vector v;  // the vector to traverse
  nat i;           // current position
public:
  vector_iterator_rep (const Vector& v2, nat i2= 0): v (v2), i (0) {}
protected:
  bool is_busy () { return i < N(v); }
  void advance () { i++; }
  C current () { return v[i]; }
  iterator_rep<C>* clone () { return new vector_iterator_rep (v, i); }
};

TMPL iterator<C>
iterate (const Vector& v) {
  return iterator<C> (new vector_iterator_rep<C,V> (v));
}

/******************************************************************************
* Hashing and generic output routines for iterators and vectors
******************************************************************************/

TMPL syntactic
flatten (const Vector& v) {
  if (is_a_scalar (v)) return flatten (v.scalar());
  nat i, n= N(v);
  vector<syntactic> r= fill <syntactic> (n);
  for (i=0; i<n; i++)
    r[i]= flatten (v[i]);
  return apply (GEN_SQTUPLE, r);
}

PRINT_SUGAR(TMPL,Vector);

TMPL
struct serialize_vector_helper {
  template<typename M> static void write (M& out, const Vector& v) {
    raw_write_length (out, N(v));
    for (nat i=0; i<N(v); i++) binary_write (out, v[i]); }
  template<typename M> static void read (M& in, Vector& v) {
    nat n;
    raw_read_length (in, n);
    nat l= aligned_size<C,V> (n);
    C* a= mmx_new<C> (l);
    for (nat i=0; i<N(v); i++) binary_read (in, a[i]);
    v= Vector (a, n, l); }
};

#define Serialize_vector_helper serialize_vector_helper<C,V>
STYPE_TO_TYPE(TMPL,binary_variant,Vector,Serialize_vector_helper );
#undef Serialize_vector_helper

/******************************************************************************
* Extracting ranges of elements and appending new elements
******************************************************************************/

TMPL Vector
range (const Vector& v, nat start, nat end) {
  VERIFY(end >= start, "bad range");
  nat n= end - start;
  if (is_a_scalar (v)) return Vector (v.scalar(), n);
  nat l= aligned_size<C,V> (n);
  C* a= mmx_new<C> (l);
  V::copy (a, seg (v) + start, n);
  return Vector (a, n, l);
}

TMPL Vector
extract_mod (const Vector& v, nat k, nat p) {
  nat n= (N(v) - k + p - 1) / p;
  nat l= aligned_size<C,V> (n);
  C* r= mmx_new<C> (l);
  for (nat i=0; i<n; i++) r[i]= v[i*p+k];
  return Vector (r, n, l);
}

TMPL inline C car (const Vector& v) { return v[0]; }
TMPL inline Vector cdr (const Vector& v) { return range (v, 1, N(v)); }

TMPL Vector
append (const Vector& v, const C& c) {
  ASSERT (is_non_scalar (v), "non-scalar vector expected");
  nat n= N(v);
  nat l= aligned_size<C,V> (n+1);
  C* a= mmx_new<C> (l);
  a[n]= c;
  V::copy (a, seg (v), n);
  return Vector (a, n+1, l);
}

TMPL Vector
append (const Vector& v, const Vector& w) {
  ASSERT (is_non_scalar (v) && is_non_scalar (w),
	  "non-scalar vectors expected");
  nat n= N(v), p= N(w);
  nat l= aligned_size<C,V> (n+p);
  C* a= mmx_new<C> (l);
  V::copy (a, seg (v), n);
  V::copy (a+n, seg (w), p);
  return Vector (a, n+p, l);
}

TMPL Vector
cons (const C& c, const Vector& v) {
  ASSERT (is_non_scalar (v), "non-scalar vector expected");
  nat n= N(v);
  nat l= aligned_size<C,V> (n+1);
  C* a= mmx_new<C> (l);
  a[0]= c;
  V::copy (a+1, seg (v), n);
  return Vector (a, n+1, l);
}

TMPL Vector&
Vector::operator << (const C& append) {
  ASSERT (is_non_scalar (*this), "non-scalar vector expected");
  secure ();
  nat n= rep->n;
  rep->resize (n+1);
  rep->a[n]= append;
  return *this;
}

TMPL Vector&
Vector::operator << (const Vector& w) {
  ASSERT (is_non_scalar (*this) && is_non_scalar (w),
	  "non-scalar vectors expected");
  secure ();
  nat n= rep->n, p= N(w);
  rep->resize (n + p);
  V::copy (rep->a + n, seg (w), p);
  return *this;
}

TMPL Vector
reverse (const Vector& v) {
  if (is_a_scalar (v)) return v;
  nat n= N(v);
  nat l= aligned_size<C,V> (n);
  C* a= mmx_new<C> (l);
  V::vec_reverse (a, seg (v), n);
  return Vector (a, n, l);
}

TMPL int
find (const Vector& v, const C& x) {
  nat i, n=N(v);
  for (i=0; i<n; i++)
    if (v[i] == x) return (int) i;
  return (int) i;
}

TMPL bool
contains (const Vector& v, const C& x) {
  return find (v, x) < ((int) N(v));
}

TMPL Vector
insert (const Vector& v, const C& x) {
  if (contains (v, x)) return v;
  else return append (v, x);
}

/******************************************************************************
* Routines on dense vectors
******************************************************************************/

template<typename Op, typename C, typename V> nat
unary_hash (const vector<C,V>& v) {
  register nat i, h= 12345, n= N(v);
  for (i=0; i<n; i++)
    h= (h<<1) ^ (h<<5) ^ (h>>27) ^ Op::op (v[i]);
  return h;
}

template<typename Op, typename C, typename V>
vector<Unary_return_type(Op,C),V>
unary_map (const vector<C,V>& v) {
  typedef Unary_return_type(Op,C) T;
  if (is_a_scalar (v)) return vector<T,V> (Op::op (v.scalar()));
  nat n= N(v);
  nat l= aligned_size<T,V> (n);
  T* r= mmx_new<T> (l);
  V::template vec_unary<Op> (r, seg (v), n);
  return vector<T,V> (r, n, l);
}

template<typename Op, typename C1, typename C2, typename V>
vector<Binary_return_type(Op,C1,C2),V>
binary_map (const vector<C1,V>& v, const vector<C2,V>& w) {
  typedef Binary_return_type(Op,C1,C2) T;
  if (is_a_scalar (v) || is_a_scalar (w)) {
    if (is_non_scalar (v)) return binary_map<Op> (v, extend (w, v));
    if (is_non_scalar (w)) return binary_map<Op> (extend (v, w), w);
    return vector<T,V> (Op::op (v.scalar(), w.scalar()));
  }
  nat n= N(v);
  ASSERT (N(w) == n, "lengths don't match");
  nat l= aligned_size<T,V> (n);
  T* r= mmx_new<T> (l);
  V::template vec_binary<Op> (r, seg (v), seg (w), n);
  return vector<T,V> (r, n, l);
}

template<typename Op, typename C, typename V, typename X>
vector<Binary_return_type(Op,C,X),V>
binary_map_scalar (const vector<C,V>& v, const X& x) {
  typedef Binary_return_type(Op,C,X) T;
  if (is_a_scalar (v)) return vector<T,V> (Op::op (v.scalar(), x));
  nat n= N(v);
  nat l= aligned_size<T,V> (n);
  T* r= mmx_new<T> (l);
  V::template vec_binary_scalar<Op> (r, seg (v), x, n);
  return vector<T,V> (r, n, l);
}

template<typename Op, typename T, typename C, typename V> vector<T,V>&
unary_update (vector<T,V>& v, const vector<C,V>& w) {
  if (is_a_scalar (v) || is_a_scalar (w)) {
    if (is_non_scalar (v))
      return unary_update<Op> (v, extend (w, v));
    else if (is_non_scalar (w))
      v= extend (v, w);
    else {
      Op::in_place (v.scalar(), w.scalar());
      return v;
    }
  }
  nat n= N(v);
  ASSERT (N(w) == n, "lengths don't match");
  V::template vec_unary<Op> (seg (v), seg (w), n);
  return v;
}

template<typename Op, typename T, typename V, typename X> vector<T,V>&
unary_update_scalar (vector<T,V>& v, const X& x) {
  if (is_a_scalar (v)) {
    Op::in_place (v.scalar(), x);
    return v;
  }
  nat n= N(v);
  V::template vec_unary_scalar<Op> (seg (v), x, n);
  return v;
}

template<typename Op, typename C1, typename C2, typename V> bool
binary_test (const vector<C1,V>& v, const vector<C2,V>& w) {
  if (is_a_scalar (v) || is_a_scalar (w)) {
    if (is_non_scalar (v)) return binary_test<Op> (v, extend (w, v));
    if (is_non_scalar (w)) return binary_test<Op> (extend (v, w), w);
    return Op::op (v.scalar(), w.scalar());
  }
  nat n= N(v);
  if (N(w) != n) return false;
  return V::template vec_binary_test<Op> (seg (v), seg (w), n);
}

template<typename Op, typename C, typename V, typename X> bool
binary_test_scalar (const vector<C,V>& v, const X& c) {
  if (is_a_scalar (v)) return Op::op (v.scalar(), c);
  return V::template vec_binary_test_scalar<Op> (seg (v), c, N(v));
}

template<typename Op, typename C, typename V> C
big (const vector<C,V>& v) {
  if (is_a_scalar (v)) {
    ASSERT (v.scalar() == Op::template neutral<C> (),
	    "infinite operation on non-neutral element");
    return v.scalar();
  }
  return V::template vec_unary_big<Op> (seg (v), N(v));
}

/******************************************************************************
* Instantiations of the abstract routines
******************************************************************************/

TMPL Vector copy (const Vector& v) {
  return unary_map<id_op> (v); }
TMPL Vector duplicate (const Vector& v) {
  return unary_map<duplicate_op> (v); }
TMPL Vector operator - (const Vector& v) {
  return unary_map<neg_op> (v); }
TMPL Vector operator + (const Vector& v, const Vector& w) {
  return binary_map<add_op> (v, w); }
TMPL Vector operator + (const C& v, const Vector& w) {
  return binary_map<add_op> (Vector (v), w); }
template<typename V> Vector_int 
operator + (const int& v, const Vector_int& w) {
  return binary_map<add_op> (Vector_int (v), w); }
TMPL Vector operator + (const Vector& w, const C& v) {
  return binary_map<add_op> (w, Vector (v)); }
template<typename V> Vector_int
operator + (const Vector_int& w, const int& v) {
  return binary_map<add_op> (w, Vector_int (v)); }
TMPL Vector operator - (const Vector& v, const Vector& w) {
  return binary_map<sub_op> (v, w); }
TMPL Vector operator - (const Vector& v, const C& w) {
  return binary_map<sub_op> (v, Vector (w)); }
template<typename V> Vector_int
operator - (const Vector_int& v, const int& w) {
  return binary_map<sub_op> (v, Vector_int (w)); }
TMPL Vector operator - (const C& v, const Vector& w) {
  return binary_map<sub_op> (Vector (v), w); }
template<typename V> Vector_int
operator - (const int& v, const Vector_int& w) {
  return binary_map<sub_op> (Vector_int (v), w); }
TMPL Vector operator * (const Vector& v, const Vector& w) {
  return binary_map<mul_op> (v, w); }
TMPL Vector operator / (const Vector& v, const Vector& w) {
  return binary_map<div_op> (v, w); }
TMPL Vector quo (const Vector& v, const Vector& w) {
  return binary_map<quo_op> (v, w); }
TMPL Vector rem (const Vector& v, const Vector& w) {
  return binary_map<rem_op> (v, w); }
TMPL Vector operator * (const Vector& v, const C& c) {
  return binary_map_scalar<rmul_op> (v, c); }
template<typename V> Vector_int
operator * (const Vector_int& v, const int& c) {
  return binary_map_scalar<rmul_op> (v, c); }
TMPL Vector operator * (const C& c, const Vector& v) {
  return binary_map_scalar<lmul_op> (v, c); }
template<typename V> Vector_int
operator * (const int& c, const Vector_int& v) {
  return binary_map_scalar<lmul_op> (v, c); }
TMPL Vector operator / (const Vector& v, const C& c) {
  return binary_map_scalar<rdiv_op> (v, c); }
template<typename V> Vector_int
operator / (const Vector_int& v, const int& c) {
  return binary_map_scalar<rdiv_op> (v, c); }
TMPL Vector operator / (const C& c, const Vector& v) {
  return binary_map_scalar<ldiv_op> (v, c); }
template<typename V> Vector_int
operator / (const int& c, const Vector_int& v) {
  return binary_map_scalar<ldiv_op> (v, c); }
TMPL Vector quo (const Vector& v, const C& c) {
  return binary_map_scalar<rquo_op> (v, c); }
TMPL Vector rem (const Vector& v, const C& c) {
  return binary_map_scalar<rrem_op> (v, c); }
TMPL Vector& operator += (Vector& v, const Vector& w) {
  return unary_update<add_op> (v, w); }
TMPL Vector& operator -= (Vector& v, const Vector& w) {
  return unary_update<sub_op> (v, w); }
TMPL Vector& operator *= (Vector& v, const Vector& w) {
  return unary_update<mul_op> (v, w); }
TMPL Vector& operator /= (Vector& v, const Vector& w) {
  return unary_update<div_op> (v, w); }
TMPL Vector& operator *= (Vector& v, const C& c) {
  return unary_update_scalar<rmul_op> (v, c); }
TMPL Vector& operator /= (Vector& v, const C& c) {
  return unary_update_scalar<rdiv_op> (v, c); }
TMPL Vector inf (const Vector& v, const Vector& w) {
  return binary_map<inf_op> (v, w); }
TMPL Vector sup (const Vector& v, const Vector& w) {
  return binary_map<sup_op> (v, w); }
TMPL bool operator <= (const Vector& v, const Vector& w) {
  return binary_test<lesseq_op> (v, w); }
TMPL bool operator >= (const Vector& v, const Vector& w) {
  return binary_test<gtreq_op> (v, w); }
TMPL bool operator == (const Vector& v, const C& c) {
  return binary_test_scalar<equal_op> (v, c); }
template<typename V> bool
operator == (const Vector_int& v, const int& c) {
  return binary_test_scalar<equal_op> (v, c); }
TMPL bool operator <= (const Vector& v, const C& c) {
  return binary_test_scalar<lesseq_op> (v, c); }
template<typename V> bool
operator <= (const Vector_int& v, const int& c) {
  return binary_test_scalar<lesseq_op> (v, c); }
TMPL bool operator >= (const Vector& v, const C& c) {
  return binary_test_scalar<gtreq_op> (v, c); }
template<typename V> bool
operator >= (const Vector_int& v, const int& c) {
  return binary_test_scalar<gtreq_op> (v, c); }

TRUE_IDENTITY_OP_SUGAR(TMPL,Vector)
SOFT_IDENTITY_OP_SUGAR(TMPL,Vector)
STRICT_COMPARE_SUGAR(TMPL,Vector)
ARITH_INT_SUGAR(TMPL,Vector,C);
EQUAL_INT_SUGAR(TMPL,Vector,Vector); 	 
COMPARE_INT_SUGAR(TMPL,Vector,Vector);

TMPL Vector sqrt (const Vector& v) { return unary_map<sqrt_op> (v); }
TMPL Vector exp (const Vector& v) { return unary_map<exp_op> (v); }
TMPL Vector log (const Vector& v) { return unary_map<log_op> (v); }
TMPL Vector cos (const Vector& v) { return unary_map<cos_op> (v); }
TMPL Vector sin (const Vector& v) { return unary_map<sin_op> (v); }
TMPL Vector tan (const Vector& v) { return unary_map<tan_op> (v); }
TMPL Vector acos (const Vector& v) { return unary_map<acos_op> (v); }
TMPL Vector asin (const Vector& v) { return unary_map<asin_op> (v); }
TMPL Vector atan (const Vector& v) { return unary_map<atan_op> (v); }
TMPL Vector pow (const Vector& v, const Vector& w) {
  return binary_map<pow_op> (v, w); }

TMPL Vector derive (const Vector& v) {
  return unary_map<derive_op> (v); }
TMPL Vector integrate (const Vector& v) {
  return unary_map<integrate_op> (v); }
template<typename C,typename V,typename X> Vector
derive (const Vector& v, const X& x) {
  return binary_map_scalar<derive_op> (v, x); }
template<typename C,typename V,typename X> Vector
integrate (const Vector& v, const X& x) {
  return binary_map_scalar<integrate_op> (v, x); }
TMPL Abs_vector abs (const Vector& v) {
  return unary_map<abs_op> (v); }

TMPL inline C big_add (const Vector& v) { return big<add_op> (v); }
TMPL inline C big_mul (const Vector& v) { return big<mul_op> (v); }
TMPL inline C big_or  (const Vector& v) { return big< or_op> (v); }
TMPL inline C big_and (const Vector& v) { return big<and_op> (v); }
TMPL inline C big_min (const Vector& v) { return big<min_op> (v); }
TMPL inline C big_max (const Vector& v) { return big<max_op> (v); }
TMPL inline C big_inf (const Vector& v) { return big<inf_op> (v); }
TMPL inline C big_sup (const Vector& v) { return big<sup_op> (v); }

/******************************************************************************
* Implementation of some routines on operators
******************************************************************************/

template<typename C> inline C
trig_op::diff_op (const C& me, const C& x) {
  return from_vector (vec (-derive (access (x, 0)) * access (me, 1),
			   derive (access (x, 0)) * access (me, 0)));
}

template<typename C> inline C
tan_op::def (const C& me, const C& f) {
  (void) me; const vector<C> v= cos_sin (f); return v[1] / v[0];
}

template<typename C> inline C
solve_vector_lde_op::diff_op (const C& me, const C& x) {
  typedef As_vector_type (C) V;
  typedef Scalar_type (V) K;
  const V w= as_vector (me);
  const V v= as_vector (x);
  nat i, n= N(v);
  K sum (0);
  V d (K (0), n);
  for (i=0; i<n; i++) {
    if (i==0) sum= v[i] * w[i];
    else sum= sum + v[i] * w[i];
    if (i<n-1) d[i]= w[i+1];
    else d[i]= sum;
  }
  return from_vector (d);
}

/******************************************************************************
* Further operations
******************************************************************************/

template<typename C> vector<C> cos_sin (const C& x) {
  return vec (cos (x), sin (x)); }
template<typename C> vector<C> trig (const vector<C>& v) {
  return vec (cos (v[0]), sin (v[0])); }

TMPL C dot (const Vector& v, const Vector& w) {
  ASSERT (is_non_scalar (v) && is_non_scalar (w),
	  "non-scalar vectors expected");
  ASSERT (N(v) == N(w), "lengths don't match");
  return V::inn_prod (seg (v), seg (w), N(v)); }

#undef TMPL_DEF
#undef TMPL
#undef Vector
#undef Vector_int
#undef Vector_rep
#undef Abs_vector
} // namespace mmx
#endif // __MMX_VECTOR_HPP
